#!/usr/bin/env zx
import 'zx/globals'
import { retry, expBackoff } from 'zx/experimental'
import {default as parseDuration} from 'parse-duration'

const configData = await fs.readFile(argv['config'] || './config.yaml', 'utf-8')
const config = YAML.parse(configData)
const wd = config['working-directory'] || 'work';
const interval = parseDuration(config.interval || '15m', 'ms')

if (!fs.existsSync(wd)) {
  echo(`Initializing ${wd}`)
  await $`mkdir ${wd}`
  cd(wd)
  await $`git init`
  await $`git remote add upstream ${config.upstream.url}`
  await $`git remote add project ${config.project.url}`
  await $`git fetch --all`
  await $`git checkout -b ${config.project.branch} project/${config.project.branch}`
  if (config.committer !== undefined) {
    await $`git config user.name ${config.committer.name}`
    await $`git config user.email ${config.committer.email}`
  }
  cd('..')
}

cd(wd)
let reportedConflictHashes
while (true) {
  await retry(30, expBackoff(), () => $`git remote prune project`.timeout('600s'))
  await retry(30, expBackoff(), () => $`git remote prune upstream`.timeout('600s'))
  await retry(30, expBackoff(), () => $`git fetch --all`.timeout('600s'))

  await $`git reset --hard project/${config.project.branch}`
  let mergeSuccess = true;
  try {
    await $`git merge upstream/${config.upstream.branch}`
  } catch (e) {
    mergeSuccess = false;
    const currentHashes = (
      (await $`git rev-parse --short project/${config.project.branch}`).stdout.trim()
      + ' and '
      + (await $`git rev-parse --short upstream/${config.upstream.branch}`).stdout.trim()
    )
    echo(chalk.red(`😱  could not merge master ${currentHashes}`))
    if (currentHashes !== reportedConflictHashes) {
      /** we saw this conflict for the first time, notify */
      reportedConflictHashes = currentHashes
      if (config?.notification?.type === 'comment-on-gitlab-merge-request') {
        const url = new URL(config.notification['merge-request-url']);
        let [projectId, mergeRequestIid] = url.pathname.substring(1).split('/merge_requests/');
        url.pathname = `/api/v4/projects/${encodeURIComponent(projectId)}/merge_requests/${mergeRequestIid}/notes`
        let body = config.notification.message
        if (body === undefined) {
          // merge again capturing the output this time
          await $`git reset --hard project/${config.project.branch}`
          const gitOutput = (await $`git merge upstream/${config.upstream.branch}`.nothrow()).toString()
          body = `Error while merging upstream changes.
\`\`\`console
$ git merge upstream/${config.upstream.branch}
${gitOutput}
\`\`\`
        `
        }
        const resp = await fetch(
          url,
          {
            method: 'post',
            body: JSON.stringify({ body }),
            headers: {
              'content-type': 'application/json',
              'PRIVATE-TOKEN': config.notification['gitlab-token'],
            }
          })
        echo(JSON.stringify(await resp.json()))
      }
    } else {
      echo(`conflict already notified`)
    }
  }
  if (mergeSuccess) {
    await retry(30, expBackoff(), () => $`git push project ${config.project.branch}`.timeout('600s'))
  }
  await sleep(interval)
}
